"""Wago modules connection."""
import logging
import socket
import struct
import time


LOG = logging.getLogger(__name__)


class Connection(object):

    """Wago connection"""

    def __init__(self, host, port, reg_addr):

        self.host = host
        self.port = port
        self.reg_addr = reg_addr

    def send(self, value):
        """Write input data to the certain register"""

        data = struct.pack('>HHHBBHH', 0, 0, 6, 1, 6, self.reg_addr, value)
        LOG.debug('Sending {0}'.format(data[5:6]))
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.socket._sock.settimeout(20)
        self.socket.connect((self.host, self.port))
        self.socket.send(data)
        time.sleep(.01)
        self.socket.close()

    def execute(self):
        """Read output data from the certain register"""

        self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.socket._sock.settimeout(20)
        self.socket.connect((self.host, self.port))
        data = struct.pack('>HHHBBHH', 0, 0, 6, 1, 3, self.reg_addr, 1)
        self.socket.send(data)
        time.sleep(.01)
        response = struct.unpack('>HHHBBBH', self.socket.recv(11))
        self.socket.close()
        if len(response) != 7:
            raise ValueError("Not enough data received")
        LOG.debug('Received {0}'.format(response[6]))
        return response[6]



from copy import deepcopy
try:
    import Queue as queue_module
except ImportError:
    import queue as queue_module
import logging
import numpy as np
from concert.imageprocessing import ramp_filter
from concert.async import threaded
from concert.imageprocessing import flat_correct as make_flat_correct
from .base import coroutine


LOG = logging.getLogger(__name__)


@coroutine
def downsize(consumer, x_slice=None, y_slice=None, z_slice=None):
    """
    downsize(consumer, x_slice=None, y_slice=None, z_slice=None)

    Downsize images in 3D and send them to *consumer*. Every argument
    is either a tuple (start, stop, step). *x_slice* operates on
    image width, *y_slice* on its height and *z_slice* on the incoming
    images, i.e. it creates the third time dimension.

    Note: the *start* index is included in the data and the *stop* index
    is excluded.
    """
    def check_and_create(sl):
        if not sl:
            sl = (0, None, 1)
        return sl

    x_start, x_stop, x_step = check_and_create(x_slice)
    y_start, y_stop, y_step = check_and_create(y_slice)
    z_start, z_stop, z_step = check_and_create(z_slice)

    i = 0
    k = 0
    while True:
        image = yield
        if z_start <= i and (not z_stop or i < z_stop):
            if k % z_step == 0:
                consumer.send(image[y_start:y_stop:y_step, x_start:x_stop:x_step])
            k += 1
        i += 1


@coroutine
def queue(consumer):
    """
    queue(consumer)

    Store the incoming data in a queue and dispatch in a separate
    thread which prevents the stalling on the "main" data stream.
    """
    item_queue = queue_module.Queue()

    @threaded
    def serve():
        while True:
            item = item_queue.get()
            consumer.send(item)
            item_queue.task_done()

    serve()

    while True:
        item = yield
        item_queue.put(deepcopy(item))


@coroutine
def average_images(consumer):
    """
    average_images(consumer)

    Average images as they come and send them to *consumer*.
    """
    average = None
    i = 0

    while True:
        data = yield
        if average is None:
            average = np.zeros_like(data, dtype=np.float32)
        average = (average * i + data) / (i + 1)
        consumer.send(average)
        i += 1


@coroutine
def sinograms(num_radiographs, consumer, sinograms_volume=None):
    """
    sinograms(num_radiographs, consumer, sinograms_volume=None)

    Convert *num_radiographs* into sinograms and send them to *consumer*.
    The sinograms are sent every time a new radiograph arrives. If there
    is more than *num_radiographs* radiographs, the sinograms are rewritten
    in a ring-buffer fashion. If *sinograms_volume* is given, it must be a 3D
    array and it is used to store the sinograms.
    """
    i = 0

    def is_compatible(radio_shape, sinos_shape):
        return radio_shape[0] == sinos_shape[0] and \
            radio_shape[1] == sinos_shape[2]

    while True:
        radiograph = yield

        if sinograms_volume is None:
            sinograms_volume = np.zeros((radiograph.shape[0], num_radiographs,
                                        radiograph.shape[1]), dtype=radiograph.dtype)
        if not is_compatible(radiograph.shape, sinograms_volume.shape):
            raise ValueError("Incompatible radiograph shape")

        sinograms_volume[:, i % num_radiographs, :] = radiograph
        consumer.send(sinograms_volume)

        i += 1


@coroutine
def flat_correct(flat, consumer, dark=None):
    """
    flat_correct(flat, consumer, dark=None)

    Flat correcting corounte, which takes a *flat* field, a *dark* field (if
    given), calculates a flat corrected radiograph and forwards it to
    *consumer*.
    """
    while True:
        radio = yield
        consumer.send(make_flat_correct(radio, flat, dark=dark))


@coroutine
def absorptivity(consumer):
    r"""
    absorptivity(consumer)

    Get the absorptivity from a flat corrected stream of images.  The intensity
    after the object is defined as :math:`I = I_0 \cdot e^{-\mu t}` and we
    extract the absorptivity :math:`\mu t` from the stream of flat corrected
    images :math:`I / I_0`.
    """
    while True:
        frame = yield
        consumer.send(-np.log(frame))


@coroutine
def stall(consumer, per_shot=10, flush_at=None):
    """
    stall(consumer, per_shot=10, flush_at=None)

    Send items once enough is collected. Collect *per_shot* items and
    send them to *consumer*. The incoming data might represent a collection
    of some kind. If the last item is supposed to be sent regardless the current
    number of collected items, use *flush_at* by which you specify the collection
    size and every time the current item *counter* % *flush_at* == 0 the item
    is sent.
    """
    i = 1

    while True:
        item = yield
        if i % per_shot == 0:
            consumer.send(item)
        elif flush_at and i % flush_at == 0:
            consumer.send(item)
            i = 0

        i += 1


@coroutine
def backproject(center, consumer):
    """
    backproject(center, consumer)

    Filtered backprojection filter. The filter receives a sinogram,
    filters it and based on *center* of rotation it backprojects it.
    The slice is then sent to *consumer*.
    """
    reco = None
    angle_step = None

    def reconstruct(sinogram, center, angle_step, x_indices, y_indices):
        """Reconstruct the slice by backprojecting all the sinogram rows."""
        width = x_indices.shape[0]
        reco = np.zeros((width, width))

        for i, phi in enumerate(np.arange(sinogram.shape[0]) * angle_step):
            pos = np.sin(phi) * x_indices + np.cos(phi) * y_indices + center
            reco += sinogram[i, pos.astype(np.int)]

        return reco

    def filter_sinogram(sinogram, start, filter_ft, width):
        """High-pass 1D filtering of the sinogram rows."""
        sinogram_ft = np.fft.fft(sinogram[:, start:start + width])
        filtered = sinogram_ft * filter_ft
        return np.fft.ifft(filtered).real.astype(np.float32)

    def get_indices(sinogram, center):
        """
        Get x and y indices which will be the base for rotation. The indices
        are created around the *center* in a symmetrical way, so if there is
        more space on one side it is cut away. This way we can mask out the
        region beyond the inscribed circle of the index arrays. Then we can
        be sure the rotated indices will always fall somewhere in the slice,
        thus we don't need to cut indices which saves time.
        """
        half = min(center, sinogram.shape[1] - center)
        y_indices, x_indices = np.mgrid[-half:half, -half:half]
        mask = np.where(np.sqrt(x_indices ** 2 + y_indices ** 2) >= half)
        x_indices[mask] = 0
        y_indices[mask] = 0

        return y_indices, x_indices

    while True:
        sinogram = yield

        if reco is None:
            if center >= sinogram.shape[1]:
                template = 'Center {} must be less than sinogram width {}'
                raise ValueError(template.format(center, sinogram.shape[1]))
            y_indices, x_indices = get_indices(sinogram, center)
            width = x_indices.shape[0]
            ramp = ramp_filter(width)
            angle_step = np.pi / sinogram.shape[0]
            # We need to store the position where we crop the sinogram, that is
            # based on the old center
            start = center - width / 2
            # Since we crop the indices the new center is always in the middle
            center = width / 2

        filtered = filter_sinogram(sinogram, start, ramp, width)
        reco = reconstruct(filtered, center, angle_step, x_indices, y_indices)
        consumer.send(reco)


class PickSlice(object):

    """Pick a slice from a 3D volume."""

    def __init__(self, index):
        self.index = index

    @coroutine
    def __call__(self, consumer):
        """Pick a slice and send it to *consumer*."""
        while True:
            volume = yield
            consumer.send(volume[self.index])



"""
Each motor is associated with a :class:`~.Calibration` that maps arbitrary
real-world coordinates to devices coordinates. When a calibration is associated
with an motor, the position can be changed with :meth:`Motor.set_position` and
:meth:`Motor.move`::

    from concert.devices.motors.dummy import Motor

    motor = Motor()

    motor.position = 2 * q.mm
    motor.move(-0.5 * q.mm)

As long as an motor is moving, :meth:`Motor.stop` will stop the motion.
"""
import logging
from concert.quantities import q
from concert.async import async
from concert.base import Quantity, State, transition, AccessorNotImplementedError
from concert.devices.base import Device


LOG = logging.getLogger(__name__)


class _PositionMixin(Device):

    """Provide positional, discrete behaviour interface."""

    def __init__(self):
        super(_PositionMixin, self).__init__()

    @async
    def move(self, delta):
        """
        move(delta)

        Move motor by *delta* user units."""
        self.position += delta

    @async
    @transition(source=['hard-limit', 'moving'], target='standby')
    def stop(self):
        """
        stop()

        Stop the motion."""
        self._stop()

    @async
    @transition(source='*', target='standby', immediate='moving')
    def home(self):
        """
        home()

        Home motor.
        """
        self._home()

    def _home(self):
        raise AccessorNotImplementedError

    def _stop(self):
        raise AccessorNotImplementedError


class LinearMotor(_PositionMixin):

    """
    One-dimensional linear motor.

    .. attribute:: position

        Position of the motor in length units.
    """

    def __init__(self):
        super(LinearMotor, self).__init__()

    def check_state(self):
        raise NotImplementedError

    state = State(default='standby')

    position = Quantity(q.m, help="Position",
                        transition=transition(source=['hard-limit', 'standby'],
                                              target=['hard-limit', 'standby'],
                                              immediate='moving', check=check_state))


class ContinuousLinearMotor(LinearMotor):

    """
    One-dimensional linear motor with adjustable velocity.

    .. attribute:: velocity

        Current velocity in length per time unit.
    """

    def __init__(self):
        super(ContinuousLinearMotor, self).__init__()

    def check_state(self):
        raise NotImplementedError

    state = State(default='standby')

    velocity = Quantity(q.m / q.s, help="Linear velocity",
                        transition=transition(source=['hard-limit', 'standby', 'moving'],
                                              target=['moving', 'standby'],
                                              check=check_state))


class RotationMotor(_PositionMixin):

    """
    One-dimensional rotational motor.

    .. attribute:: position

        Position of the motor in angular units.
    """

    state = State(default='standby')

    def check_state(self):
        raise NotImplementedError

    position = Quantity(q.deg, help="Angular position",
                        transition=transition(source=['hard-limit', 'standby'],
                                              target=['hard-limit', 'standby'],
                                              immediate='moving', check=check_state))

    def __init__(self):
        super(RotationMotor, self).__init__()


class ContinuousRotationMotor(RotationMotor):

    """
    One-dimensional rotational motor with adjustable velocity.

    .. attribute:: velocity

        Current velocity in angle per time unit.
    """

    def __init__(self):
        super(ContinuousRotationMotor, self).__init__()

    def check_state(self):
        raise NotImplementedError

    state = State(default='standby')

    velocity = Quantity(q.deg / q.s, help="Angular velocity",
                        transition=transition(source=['hard-limit', 'standby', 'moving'],
                                              target=['moving', 'standby'], check=check_state))



"""WAGO-I/O-SYSTEM 750"""

from concert.devices.io import base
from concert.networking.wago import Connection


class IO(base.IO):

    """Input/output wago system implementation"""

    def __init__(self, host, port):
        super(IO, self).__init__()
        self._host = host
        self._port = port
        self._ports = range(256)

    def _write_port(self, port, value):
        """Set input value"""
        connection = Connection(self._host, self._port, port)
        connection.send(value)

    def _read_port(self, port):
        """Read output value"""
        connection = Connection(self._host, self._port, port)
        response = connection.execute()
        return int(response)



"""Motor Dummy."""
import random
from concert.base import HardLimitError
from concert.quantities import q
from concert.devices.motors import base


class _PositionMixin(object):
    def __init__(self):
        self.lower = -100 * q.mm
        self.upper = +100 * q.mm
        self._position = random.uniform(self.lower, self.upper)

    def _set_position(self, position):
        if position < self.lower:
            self._position = self.lower
        elif position > self.upper:
            self._position = self.upper
        else:
            self._position = position

    def _get_position(self):
        return self._position

    def _stop(self):
        pass


class _ContinuousMixin(object):
    def __init__(self):
        super(_ContinuousMixin, self).__init__()
        self.velocity_lower = -100 * q.mm / q.s
        self.velocity_upper = 100 * q.mm / q.s
        self._velocity = 0 * q.mm / q.s

    def _set_velocity(self, velocity):
        if velocity < self.velocity_lower:
            self._velocity = self.velocity_lower
        elif velocity > self.velocity_upper:
            self._velocity = self.velocity_upper
        else:
            self._velocity = velocity

    def _get_velocity(self):
        return self._velocity


class LinearMotor(base.LinearMotor, _PositionMixin):

    """A linear step motor dummy."""

    def __init__(self, position=None):
        super(LinearMotor, self).__init__()
        _PositionMixin.__init__(self)

        if position:
            self._position = position

    def check_state(self):
        if self._position > self.lower and self._position < self.upper:
            return 'standby'

        raise HardLimitError('hard-limit')


class ContinuousLinearMotor(LinearMotor, base.ContinuousLinearMotor, _ContinuousMixin):

    """A continuous linear motor dummy."""

    def __init__(self):
        super(ContinuousLinearMotor, self).__init__()
        _ContinuousMixin.__init__(self)

    def check_state(self):
        if self.velocity.magnitude != 0:
            return 'moving'

        return LinearMotor.check_state(self)

    def _stop(self):
        self._velocity = 0 * q.mm / q.s


class RotationMotor(base.RotationMotor, _PositionMixin):

    """A rotational step motor dummy."""

    def __init__(self):
        super(RotationMotor, self).__init__()
        _PositionMixin.__init__(self)
        self.lower = -float("Inf") * q.deg
        self.upper = float("Inf") * q.deg
        self._position = self._position.magnitude * q.deg

    def check_state(self):
        return 'standby'


class ContinuousRotationMotor(RotationMotor,
                              base.ContinuousRotationMotor,
                              _ContinuousMixin):

    """A continuous rotational step motor dummy."""

    def __init__(self):
        super(ContinuousRotationMotor, self).__init__()
        _ContinuousMixin.__init__(self)
        self.velocity_lower = self.velocity_lower.magnitude * q.deg / q.s
        self.velocity_upper = self.velocity_upper.magnitude * q.deg / q.s
        self._velocity = self._velocity.magnitude * q.deg / q.s

    def check_state(self):
        if self.velocity != 0:
            return 'moving'

        return RotationMotor.check_state(self)

    def _stop(self):
        self._velocity = 0 * q.deg / q.s



from concert.quantities import q
from concert import optimization
from concert.tests import slow, assert_almost_equal, TestCase
from concert.optimization import optimize_parameter
from concert.devices.motors.dummy import LinearMotor


class TestOptimizers(TestCase):

    def setUp(self):
        super(TestOptimizers, self).setUp()
        self.algorithms = [optimization.halver, optimization.down_hill,
                           optimization.powell,
                           optimization.nonlinear_conjugate,
                           optimization.bfgs,
                           optimization.least_squares]
        self.center = 1.0 * q.mm
        self.motor = LinearMotor(position=0 * q.mm)
        self.motor.lower = -float('Inf') * q.mm
        self.motor.upper = float('Inf') * q.mm

    def feedback(self):
        return (self.motor.position.to(q.mm).magnitude -
                self.center.to(q.mm).magnitude) ** 2

    def check(self):
        assert_almost_equal(self.motor.position.to(q.mm),
                            self.center.to(q.mm), 1e-2)

    def optimize(self, algorithm):
        optimize_parameter(self.motor["position"], self.feedback,
                           self.motor.position, algorithm).join()

    @slow
    def test_algorithms(self):
        for i in range(len(self.algorithms)):
            self.motor.position = 0 * q.mm
            self.optimize(self.algorithms[i])
            self.check()



import time
import random
from concert.devices.dummy import DummyDevice
from concert.async import async, wait, KillException, HAVE_GEVENT
from concert.tests import slow, TestCase


@async
def func():
    pass


@async
def bad_func():
    raise RuntimeError


class TestAsync(TestCase):

    def setUp(self):
        super(TestAsync, self).setUp()
        self.device = DummyDevice()

    @slow
    def test_wait(self):
        @async
        def long_func():
            time.sleep(random.random() / 50.)

        futs = []
        for i in range(10):
            futs.append(long_func())

        wait(futs)

        for future in futs:
            self.assertTrue(future.done(), "Not all futures finished.")

    def test_exceptions(self):
        with self.assertRaises(TypeError):
            func(0).join()

        with self.assertRaises(RuntimeError):
            bad_func().join()

    def test_kill(self):
        d = {'killed': False}

        @async
        def long_op(d):
            try:
                time.sleep(10)
            except KillException:
                d['killed'] = True

        future = long_op(d)
        time.sleep(0)
        future.kill()

        if HAVE_GEVENT:
            self.assertTrue(d['killed'])



"""Shutter Dummy."""
from concert.devices.shutters import base


class Shutter(base.Shutter):

    """A dummy shutter that can be opened and closed."""

    def __init__(self):
        super(Shutter, self).__init__()
        self._dummy_state = 'open'

    def _open(self):
        self._dummy_state = 'open'

    def _close(self):
        self._dummy_state = 'closed'



"""Storage implementations."""
import os
import logging
from concert.ext import tifffile


LOG = logging.getLogger(__name__)


def read_image(file_name):
    """
    Read image from file with *file_name*. The file type is detected
    automatically.
    """
    ext = os.path.splitext(file_name)[1]
    if ext in READERS:
        return READERS[ext](file_name)

    raise ValueError("Unsupported file type")


def read_tiff(file_name):
    """Read tiff file from disk by :py:mod:`tifffile` module."""
    return tifffile.imread(file_name)


READERS = {".tif": read_tiff,
           ".tiff": read_tiff}


def write_tiff(file_name_prefix, data):
    """
    The default TIFF writer which uses :py:mod:`tifffile` module.
    Return the written file name.
    """
    file_name = file_name_prefix + ".tif"
    tifffile.imsave(file_name, data)

    return file_name


def write_libtiff(file_name_prefix, data):
    """Write a TIFF file using pylibtiff. Return the written file name."""
    from libtiff import TIFF

    file_name = file_name_prefix + ".tif"
    tiff_file = TIFF.open(file_name, "w")
    try:
        tiff_file.write_image(data)
    finally:
        tiff_file.close()

    return file_name


def create_directory(directory, rights=0o0750):
    """Create *directory* and all paths along the way if necessary."""
    if not os.path.exists(directory):
        LOG.debug("Creating directory {}".format(directory))
        os.makedirs(directory, rights)



"""Connection to Connection controllers."""
import logging
from concert.networking.base import SocketConnection


LOG = logging.getLogger(__name__)


class Connection(SocketConnection):

    """Aerotech socket connection."""
    EOS_CHAR = "\n"  # string termination character
    ACK_CHAR = "%"  # acknowledge
    NAK_CHAR = "!"  # not acknowledge (wrong parameters, etc.)
    FAULT_CHAR = "#"  # task fault

    def __init__(self, host, port):
        super(Connection, self).__init__(host, port, return_sequence=Connection.EOS_CHAR)

    @classmethod
    def _interpret_response(cls, hle_response):
        if not hle_response:
            raise ValueError("Not enough data received")
        if (hle_response[0] == Connection.ACK_CHAR):
            # return the data
            res = hle_response[1:]
            LOG.debug("Interpreted response {0}.".format(res))
            return res
        if (hle_response[0] == Connection.NAK_CHAR):
            LOG.warn(hle_response)
            raise ValueError("Invalid command or parameter")
        if (hle_response[0] == Connection.FAULT_CHAR):
            raise RuntimeError("Controller task error.")

    def recv(self):
        """Return properly interpreted answer from the controller."""
        return self._interpret_response(super(Connection, self).recv())



"""Adventurer scales."""

from concert.devices.scales import base
from concert.devices.base import Device
from concert.networking import SocketConnection
from concert.quantities import q
from concert.devices.scales.base import WeightError


class ARRW60(base.TarableScales):

    """The ARRW60 model of Adventurer scales."""
    ERROR = "error"
    OK = "ok"

    def __init__(self, host, port):
        super(ARRW60, self).__init__()
        self._connection = SocketConnection(host, port, "\r\n")
        self._states = self._states.union(set([ARRW60.OK, ARRW60.ERROR]))

    def _execute(self, cmd):
        result = self._connection.execute(cmd)
        if "OK!" not in result:
            raise ValueError("Bad command or value")

    def _tare(self):
        self._execute("T")

    def _get_weight(self):
        """
        Get weight from the scales. The command does not return until
        a balanced position is found, thus it can time out.
        """
        res = self._connection.execute("P")
        if "Err8.4" in res:
            self._set_state(ARRW60.ERROR)
            raise WeightError("More than maximum weight loaded")
        else:
            if self.state == ARRW60.ERROR or self.state == Device.NA:
                # Clear the error from before or set OK for the first time
                self._set_state(ARRW60.OK)
            # The returned string contains units
            res = q.Quantity(res).to(q.g)

        return float(res.magnitude) * q.counts



"""
Motors based on the CompactRIO controller.
"""
from concert.quantities import q
from concert.networking import SocketConnection
from concert.devices.motors import base


DEFAULT_CRIO_HOST = 'cRIO9074-Motion.ipe.kit.edu'
DEFAULT_CRIO_PORT_LINEAR = 6342
DEFAULT_CRIO_PORT_ROTATIONAL = 6340


class LinearMotor(base.LinearMotor):

    """A linear motor that moves in two directions."""

    def __init__(self, host=DEFAULT_CRIO_HOST, port=DEFAULT_CRIO_PORT_LINEAR):
        super(LinearMotor, self).__init__()
        self._connection = SocketConnection(host, port, '\r\n')
        self._position = 0 * q.mm
        self._to_device_scale = 50000
        self['position'].lower = 0 * q.mm
        self['position'].upper = 2 * q.mm

    def _get_position(self):
        return self._position / self._to_device_scale

    def _set_position(self, position):
        self._position = position.to(q.mm) * self._to_device_scale
        self._connection.execute('lin {} 5000'.format(int(self._position)))

    def _stop(self):
        self._connection.execute('mod stop')


class RotationMotor(base.ContinuousRotationMotor):

    """A rotational motor."""

    def __init__(self, host=DEFAULT_CRIO_HOST, port=DEFAULT_CRIO_PORT_ROTATIONAL):
        super(RotationMotor, self).__init__()
        self._connection = SocketConnection(host, port, '\r\n')
        self._position = 0 * q.deg
        self._velocity = 0 * q.deg / q.s
        self._to_device_scale = 2

    def _in_hard_limit(self):
        return False

    def _in_velocity_hard_limit(self):
        return False

    def _get_position(self):
        return self._position / self._to_device_scale

    def _set_position(self, position):
        self._position = position * self._to_device_scale
        self._connection.execute('rot {} 500'.format(int(self._position)))

    def _get_velocity(self):
        return 0 * q.deg / q.s

    def _set_velocity(self, velocity):
        self._velocity = velocity * self._to_device_scale
        self._connection.execute('mod {}'.format(int(self._velocity)))

    def _stop(self):
        self._connection.execute('mod stop')



from concert import optimization
from concert.base import HardLimitError
from concert.quantities import q
from concert.tests import slow, assert_almost_equal, TestCase
from concert.measures import DummyGradientMeasure
from concert.devices.motors.dummy import LinearMotor
from concert.optimization import optimize_parameter


class TestDummyFocusingWithSoftLimits(TestCase):

    def setUp(self):
        super(TestDummyFocusingWithSoftLimits, self).setUp()
        self.motor = LinearMotor(position=50 * q.mm)
        self.motor['position'].lower = 25 * q.mm
        self.motor['position'].upper = 75 * q.mm
        self.halver_kwargs = {"initial_step": 10 * q.mm,
                              "max_iterations": 1000}

    @slow
    def test_maximum_out_of_soft_limits_right(self):
        feedback = DummyGradientMeasure(self.motor['position'], 80 * q.mm)
        optimize_parameter(self.motor["position"], lambda: - feedback(),
                           self.motor.position, optimization.halver,
                           alg_kwargs=self.halver_kwargs).join()
        assert_almost_equal(self.motor.position, 75 * q.mm)

    @slow
    def test_maximum_out_of_soft_limits_left(self):
        feedback = DummyGradientMeasure(self.motor['position'], 20 * q.mm)
        optimize_parameter(self.motor["position"], lambda: - feedback(),
                           self.motor.position, optimization.halver,
                           alg_kwargs=self.halver_kwargs).join()
        assert_almost_equal(self.motor.position, 25 * q.mm)


class TestDummyFocusing(TestCase):

    def setUp(self):
        super(TestDummyFocusing, self).setUp()
        self.motor = LinearMotor()
        self.motor.position = 0 * q.mm
        self.feedback = DummyGradientMeasure(self.motor['position'],
                                             18.75 * q.mm)
        self.epsilon = 1 * q.um
        self.position_eps = 1e-1 * q.mm
        self.gradient_cmp_eps = 1e-1
        self.halver_kwargs = {"initial_step": 10 * q.mm,
                              "max_iterations": 3000}

    def run_optimization(self, initial_position=1 * q.mm):
        optimize_parameter(self.motor["position"], lambda: - self.feedback(),
                           initial_position, optimization.halver,
                           alg_kwargs=self.halver_kwargs).join()

    def check(self, other):
        assert_almost_equal(self.motor.position, other, 0.1)

    @slow
    def test_maximum_in_limits(self):
        self.run_optimization()
        self.check(self.feedback.max_position)

    @slow
    def test_huge_step_in_limits(self):
        self.halver_kwargs["initial_step"] = 1000 * q.mm

        with self.assertRaises(HardLimitError):
            self.run_optimization()

    @slow
    def test_maximum_out_of_limits_right(self):
        self.feedback.max_position = (self.motor.upper + 50 * q.mm)

        with self.assertRaises(HardLimitError):
            self.run_optimization()

    @slow
    def test_maximum_out_of_limits_left(self):
        self.feedback.max_position = (self.motor.lower - 50 * q.mm)

        with self.assertRaises(HardLimitError):
            self.run_optimization()

    @slow
    def test_identical_gradients(self):
        # Some gradient level reached and then by moving to another position
        # the same level is reached again, but global gradient maximum
        # is not at this motor position.
        self.halver_kwargs["initial_step"] = 10 * q.mm
        self.run_optimization(initial_position=-0.00001 * q.mm)
        self.check(self.feedback.max_position)



import numpy as np
from nose.plugins.attrib import attr
from concert.quantities import q
from concert.devices.motors.dummy import ContinuousRotationMotor
from concert.tests import slow, TestCase
from concert.tests.util.rotationaxis import SimulationCamera
from concert.processes import scan
from concert.measures import rotation_axis


@attr('skip-travis')
class TestRotationAxisMeasure(TestCase):

    def setUp(self):
        super(TestRotationAxisMeasure, self).setUp()
        self.x_motor = ContinuousRotationMotor()
        self.y_motor = ContinuousRotationMotor()
        self.z_motor = ContinuousRotationMotor()

        # The bigger the image size, the more images we need to determine
        # the center correctly.
        self.image_source = SimulationCamera(128, self.x_motor["position"],
                                             self.y_motor["position"],
                                             self.z_motor["position"])

        # Allow 1 px misalignment in y-direction.
        self.eps = np.arctan(2.0 / self.image_source.rotation_radius) * q.rad

    def make_images(self, x_angle, z_angle, intervals=10):
        self.x_motor.position = z_angle
        self.z_motor.position = x_angle
        return scan(self.y_motor["position"],
                    self.image_source.grab, minimum=0 * q.rad,
                    maximum=2 * np.pi * q.rad,
                    intervals=intervals).result()[1]

    def align_check(self, x_angle, z_angle):
        images = self.make_images(x_angle, z_angle)
        phi, psi = rotation_axis(images)[:2]

        assert phi + x_angle < self.eps
        assert np.abs(psi) - np.abs(z_angle) < self.eps

    def center_check(self, images):
        center = rotation_axis(images)[2]

        assert np.abs(center[1] - self.image_source.ellipse_center[1]) < 2
        assert np.abs(center[0] - self.image_source.ellipse_center[0]) < 2

    @slow
    def test_out_of_fov(self):
        images = np.ones((10, self.image_source.size,
                          self.image_source.size))
        with self.assertRaises(ValueError) as ctx:
            rotation_axis(images)

        self.assertEqual("No sample tip points found.", str(ctx.exception))

    @slow
    def test_center_no_rotation(self):
        images = self.make_images(0 * q.deg, 0 * q.deg, intervals=15)
        self.center_check(images)

    @slow
    def test_center_only_x(self):
        images = self.make_images(17 * q.deg, 0 * q.deg, intervals=15)
        self.center_check(images)

    @slow
    def test_center_only_z(self):
        images = self.make_images(0 * q.deg, 11 * q.deg, intervals=15)
        self.center_check(images)

    @slow
    def test_center_positive(self):
        images = self.make_images(17 * q.deg, 11 * q.deg, intervals=15)
        self.center_check(images)

    @slow
    def test_center_negative_positive(self):
        images = self.make_images(-17 * q.deg, 11 * q.deg, intervals=15)
        self.center_check(images)

    @slow
    def test_center_positive_negative(self):
        images = self.make_images(17 * q.deg, -11 * q.deg, intervals=15)
        self.center_check(images)

    @slow
    def test_center_negative(self):
        images = self.make_images(-17 * q.deg, -11 * q.deg, intervals=15)
        self.center_check(images)

    @slow
    def test_only_x(self):
        """Only misaligned laterally."""
        self.align_check(0 * q.deg, 0 * q.deg)

    @slow
    def test_only_z(self):
        """Only misaligned in the beam direction."""
        self.align_check(0 * q.deg, 11 * q.deg)

    @slow
    def test_huge_x(self):
        self.image_source.scale = (3, 0.25, 3)
        self.align_check(60 * q.deg, 11 * q.deg)

    @slow
    def test_huge_z(self):
        self.image_source.scale = (3, 0.25, 3)
        self.align_check(11 * q.deg, 60 * q.deg)

    @slow
    def test_positive(self):
        self.align_check(17 * q.deg, 11 * q.deg)

    @slow
    def test_negative_positive(self):
        self.align_check(-17 * q.deg, 11 * q.deg)

    @slow
    def test_positive_negative(self):
        self.align_check(17 * q.deg, -11 * q.deg)

    @slow
    def test_negative(self):
        self.align_check(-17 * q.deg, -11 * q.deg)



"""Imaging experiments usually conducted at synchrotrons."""
import os
import numpy as np
from concert.quantities import q
from concert.storage import write_tiff, create_directory
from concert.coroutines.sinks import write_images
from concert.experiments.base import Experiment as BaseExperiment


class Experiment(BaseExperiment):

    """
    Imaging experiment stores images acquired in acquisitions on disk
    automatically.
    """

    def __init__(self, acquisitions, directory_prefix, log=None, log_file_name="experiment.log",
                 writer=write_tiff):
        super(Experiment, self).__init__(acquisitions, directory_prefix, log=log,
                                         log_file_name=log_file_name)
        self.writer = writer
        self._writers = {}

    def acquire(self):
        """Run the experiment. Add writers to acquisitions dynamically."""
        try:
            for acq in self.acquisitions:
                self._attach_writer(acq)
                acq()
        finally:
            self._remove_writers()

    def _attach_writer(self, acquisition):
        """
        Attach a writer to an *acquisition* in order to store the images on
        disk automatically.
        """
        directory = os.path.join(self.directory, acquisition.name)
        create_directory(directory)
        prefix = os.path.join(directory, "frame_{:>05}")

        if acquisition not in self._writers:
            writer = write_images(writer=self.writer, prefix=prefix)
            # The launcher returns the original writer even if the acquisition
            # is invoked multiple times. In that case the images from every run
            # are appended to the original ones.
            launcher = lambda: writer
            acquisition.consumers.append(launcher)
            self._writers[acquisition] = launcher

    def _remove_writers(self):
        """
        Cleanup the image writing in order to be able to write images into
        the current directory during the next scan.
        """
        for acq, launcher in self._writers.items():
            acq.consumers.remove(launcher)
        self._writers = {}


def frames(num_frames, camera, callback=None):
    """
    A generator which takes *num_frames* using *camera*. *callback* is called
    after every taken frame.
    """
    if camera.state == 'recording':
        camera.stop_recording()

    camera['trigger_mode'].stash().join()
    camera.trigger_mode = camera.trigger_modes.SOFTWARE

    try:
        with camera.recording():
            for i in range(num_frames):
                camera.trigger()
                yield camera.grab()
                if callback:
                    callback()
    finally:
        camera['trigger_mode'].restore().join()


def tomo_angular_step(frame_width):
    """
    Get the angular step required for tomography so that every pixel of the frame
    rotates no more than one pixel per rotation step. *frame_width* is frame size in
    the direction perpendicular to the axis of rotation.
    """
    return np.arctan(2.0 / frame_width.magnitude) * q.rad


def tomo_projections_number(frame_width):
    """
    Get the minimum number of projections required by a tomographic scan in
    order to provide enough data points for every distance from the axis of
    rotation. The minimum angular step is
    considered to be needed smaller than one pixel in the direction
    perpendicular to the axis of rotation. The number of pixels in this
    direction is given by *frame_width*.
    """
    return int(np.ceil(np.pi / tomo_angular_step(frame_width)))


def tomo_max_speed(frame_width, frame_rate):
    """
    Get the maximum rotation speed which introduces motion blur less than one
    pixel. *frame_width* is the width of the frame in the direction
    perpendicular to the rotation and *frame_rate* defines the time required
    for recording one frame.

    _Note:_ frame rate is required instead of exposure time because the
    exposure time is usually shorter due to the camera chip readout time.
    We need to make sure that by the next exposure the sample hasn't moved
    more than one pixel from the previous frame, thus we need to take into
    account the whole frame taking procedure (exposure + readout).
    """
    return tomo_angular_step(frame_width) * frame_rate



import numpy as np
from concert.quantities import q
from concert.imageprocessing import compute_rotation_axis
from concert.tests import suppressed_logging, slow, assert_almost_equal


@slow
@suppressed_logging
def test_rotation_axis():
    """
    Test tomographic rotation axis finding. The sample is a triangle which
    is offcentered by different values.
    """
    def triangle(n, width, position, left=True):
        tr = np.zeros((width, width), dtype=np.float)
        extended = np.zeros((n, n), dtype=np.float)

        indices = np.tril_indices(width)
        tr[indices] = 1
        if not left:
            tr = tr[:, ::-1]
        extended[n / 2 - width / 2: n / 2 + width / 2, position:position + width] = tr

        return extended

    def test_axis(n, width, left_position, right_position):
        left = triangle(n, width, left_position)
        right = triangle(n, width, right_position, left=False)

        center = compute_rotation_axis(left, right)
        truth = (left_position + right_position + width) / 2.0 * q.px
        assert_almost_equal(center, truth)

    n = 128
    width = 32

    # Axis is exactly in the middle
    test_axis(n, width, n / 2 - width, n / 2)

    # Axis is to the left from the center
    test_axis(n, width, 8, n / 2 + width / 4)

    # Axis is to the right from the center
    test_axis(n, width, 8, n - width)



"""
Measures operate on some data and provide information about how \"good\"
the data is.
"""
import numpy as np
from concert.quantities import q
from concert.imageprocessing import center_of_points, needle_tips


class DummyGradientMeasure(object):

    """Gradient measure that returns a quadratic fall-off of *parameter* from
    *max_position*."""

    def __init__(self, parameter, max_position):
        self.max_position = max_position
        self._max_gradient = 1e4
        self._param = parameter

    def __call__(self):
        value = self._param.get().result()
        position = value.to(self.max_position.units).magnitude
        return self._max_gradient - \
            (position - self.max_position.magnitude) ** 2


class Area(object):

    """
    Area measures how much of the whole image is covered by an object.
    The return value is the fraction between object pixels and the whole
    area, :math:`n = object\_pixels / (width \cdot height)`.

    Parameters are:

    .. py:attribute:: func

        a callable *func(radio, \*args, \*\*kwargs)*, where
        *radio* is a radiograph to be analyzed, *args* and *kwargs*
        are additional positional and keyword arguments passed to *func*
        which returns the number of pixels covered by
        an object

    .. py:attribute:: args

        *func* arguments tuple

    .. py:attribute:: kwargs

        *func* keyword arguments dictionary

    """

    def __init__(self, func, args=None, kwargs=None):
        self.func = func
        self.args = () if args is None else args
        self.kwargs = {} if kwargs is None else kwargs

    def __call__(self, radio):
        object_pixels = self.func(radio, *self.args, **self.kwargs)

        return float(object_pixels) / (radio.shape[0] * radio.shape[1])


class SimpleArea(Area):

    """
    An area measure with object segmentation by a threshold. If two
    flat fields *flat_1* and *flat_2* are given, then the minimum
    of their subtraction :math:`min(flat\_1, flat\_2)` approximates
    the maximum grey value considered a background. The values
    from the :math:`radiograph - flat` below the minimum are
    considered an object.
    """

    def __init__(self, flat_1, flat_2):
        super(SimpleArea, self).__init__(self.object_pixels)
        self._flat_1 = flat_1
        self._flat_2 = flat_2
        self.flat_avg = None
        self.threshold = None
        self._setup()

    def object_pixels(self, radio):
        """Get object pixels from a radiograph *radio*."""
        radio = radio - self.flat_avg

        return len(np.where(radio < self.threshold)[0])

    @property
    def flats(self):
        """flat_1 used by measure"""
        return [self._flat_1, self._flat_2]

    @flats.setter
    def flats(self, flats):
        """
        Set new flat fields flat_1 and flat_2 from the tuple *flats*.
        The threshold and average flat are recomputed.
        """
        if flats[0].shape != flats[1].shape:
            raise ValueError("Flats must have the same shape")
        self._flat_1 = flats[0]
        self._flat_2 = flats[1]
        self._setup()

    def _setup(self):
        """Setup the threshold and average flat field."""
        # Threshold
        self.threshold = np.min(self._flat_1 - self._flat_2)
        # Average the two flats to suppress the noise a little
        self.flat_avg = (self._flat_1 + self._flat_2) / 2.0


def rotation_axis(images):
    """
    Determine a 3D circle normal inclination angles towards the
    :math:`y`-axis (0,1,0) and the center of the circle.

    **Algorithm**

    We exploit the fact that the projected circle is an ellipse. Thus,
    ellipse's major axis is the radius of rotation. If we set the beam
    direction to be the :math:`z`-axis :math:`(0,0,1)`, we can determine
    the :math:`xy` angle :math:`(\phi)` by the major axis angle
    of the ellipse. Furthermore, based on the ratio between minor
    and major axis, we can determine the projected circle's
    inclination from its normal, yielding the :math:`zy` angle
    :math:`(\psi)` of the circle. One then needs to rotate the the sample
    :math:`(-\phi,\pm \psi)`. The :math:`\psi` angle ambiguity comes from
    the fact that we are working with projected data.

    There are three possible situations with the ellipse fitting algorithm:

    * The sample does not move in the image sequence, it is not off-centered
        enough.
    * The ellipse degenerates to a line, which means the circle is parallel
        to the beam, thus we only need to rotate by :math:`-\phi`.
    * The projections lead to a non-degenerate ellipse. In this case we need to
        rotate around both angles.

    **Input**

    An image sequence of at least 5 images with different sample positions
    around the axis of rotation. It is highly recommended to use more images
    in order to obtain more precise results from noisy data.

    **Steps**

    The measuring procedure is done in the following steps:

    #. Sample extraction.
    #. Sample tip determination (largest Euclidean distance from the sample
        entering to the image and its boundaries).
    #. Ellipse fitting to the set of tip points.
    #. Determine :math:`\phi` and :math:`\psi` from the projected circle.

    **Assumptions**

    The procedure assumes a high-absorbing sample with conic shape. Thanks to
    such sample the segmentation procedure can be simplified to a thresholding
    technique.

    **Output**

    The measure returns a tuple (:math:`\phi`, :math:`\psi`, center).
    """
    if len(images) < 5:
        raise ValueError("At least 5 images are needed")
    params, tips = _fit_ellipse(images)
    phi, psi = _get_ellipse_angles(params, tips)
    center = _get_ellipse_center(params, tips)

    return phi, psi, center


def _get_ellipse_center(params, tips):
    """Determine the center of the ellipse from its parameters."""
    a_33 = _get_conic_submatrix(params)
    if np.linalg.det(a_33) > 1e-3:
        x_pos = (params[1] * params[4] -
                 2 * params[3] * params[2]) /\
            (4 * params[0] * params[2] - params[1] ** 2)
        y_pos = (params[3] * params[1] -
                 2 * params[0] * params[4]) /\
            (4 * params[0] * params[2] - params[1] ** 2)
    else:
        # We are not dealing with an ellipse, use center of mass.
        y_pos, x_pos = center_of_points(tips)

    # +1 for image edge-clipping compensation.
    return y_pos + 1, x_pos + 1


def _get_ellipse_angles(params, tips):
    """
    Determine the :math:`xy` :math:`(\phi)` and :math:`zy` angle
    :math:`(\psi)` of the normal of the circle defined by the axis
    of rotation. The axis is determined by fitting an ellipse
    to the sample rotation visible by different tips positions
    in images. Return a tuple :math:`(\phi`, :math:`\psi)`.
    """
    y_ind, x_ind = zip(*tips)
    x_ind = np.array(x_ind)
    y_ind = np.array(y_ind)
    d_x = x_ind.max() - x_ind.min()
    if d_x < 10:
        raise ValueError("Sample off-centering too small, " +
                         "enlarge rotation radius.")

    a_33 = _get_conic_submatrix(params)

    usv = np.linalg.svd(a_33)
    s_vec = usv[1]
    v_mat = usv[2]

    if np.linalg.det(a_33) <= 0:
        # Not an ellipse
        d_y = float(y_ind.max() - y_ind.min())
        angle = np.arctan(d_y / d_x) * q.rad
        if x_ind[y_ind.argmax()] <= x_ind[y_ind.argmin()]:
            # More than pi/4.
            angle = -angle
        phi, psi = angle, 0.0 * q.rad
    else:
        phi, psi = \
            np.arctan(v_mat[1][1] / v_mat[1][0]) * q.rad, \
            np.arcsin(np.sqrt(s_vec[1]) / np.sqrt(s_vec[0])) * q.rad

    return phi, psi


def _get_conic_submatrix(params):
    """
    Get the conic section submatrix for determining the conic section
    type. *params* are the conic section parameters A, B, C, D, E, F,
    where the conic section is :math:`Ax^2 + Bxy + Cy^2 + Dx + Ey + F = 0`.
    """
    return np.array([[params[0], params[1] / 2], [params[1] / 2, params[2]]])


def _fit_ellipse(images):
    """Ellipse fitting based on *points* and singular value
    decomposition.
    """
    tips = needle_tips(images)

    a_matrix = _construct_matrix(tips)
    v_mat = np.linalg.svd(a_matrix)[2]
    params = v_mat.T[:, -1]

    return params, tips


def _construct_matrix(points):
    """Construct a conical section matrix from data *points*."""
    y_ind, x_ind = zip(*points)
    matrix = np.empty((len(x_ind), 6), dtype=np.int)
    for i in range(len(x_ind)):
        matrix[i] = np.array([x_ind[i] ** 2, x_ind[i] * y_ind[i],
                              y_ind[i] ** 2, x_ind[i], y_ind[i], 1])

    return matrix



"""
Fiber-lite halogen light source
"""
from concert.quantities import q
from concert.devices.io.wago import IO
from concert.devices.lightsources import base


class FiberLite(base.LightSource):

    """
    The class for controlling of intensity of halogen light source by voltage
    """

    def __init__(self, host, port, wago_port):

        self._host = host
        self._port = port
        self._wago_port = wago_port
        self._intensity = 0 * q.V
        super(FiberLite, self).__init__()

    def _set_intensity(self, value):
        """
        Set the intensity by voltage for the FiberLight
        """
        connection = IO(self._host, self._port)
        connection.write_port(self._wago_port, value.magnitude * 51)
        self._intensity = value

    def _get_intensity(self):
        """
        Get the voltage from the FiberLite corresponding to current intensity
        """
        return self._intensity
